@@ -1,15 +1,15 @@
 import io
+import json
 import struct
 
-import fastavro
 from Crypto.Hash import SHA256
 from Crypto.Random import get_random_bytes
 
-from .utils import b64dec, b64enc, b64dec_if_not_bytes, b64enc_if_not_str, le_u256enc, le_u256dec, encode_tag, decode_tag, normalize_tag
+from .utils import b64dec, b64enc, b64dec_if_not_bytes, b64enc_if_not_str, encode_tag, decode_tag, normalize_tag, create_tag, tags_to_dict
 from .utils.deep_hash import deep_hash
 from .utils.ans104_signers import DEFAULT as DEFAULT_SIGNER, BY_TYPE as SIGNERS_BY_TYPE
 
-ANS104_TAGS_SCHEMA = {
+ANS104_TAGS_AVRO_SCHEMA = {
   "type": "array",
   "items": {
     "type": "record",
@@ -21,8 +21,6 @@ ANS104_TAGS_SCHEMA = {
   }  
 }
 
-ANS104_TAGS_SCHEMA_fastavro = fastavro.parse_schema(ANS104_TAGS_SCHEMA)
-
 class ANS104BundleHeader:
     def __init__(self, length_by_id = {}, version = 2):
         self.length_by_id = length_by_id
@@ -50,14 +48,14 @@ class ANS104BundleHeader:
         return 32 + len(self.length_by_id) * 64
 
     def tobytes(self):
-        return le_u256enc(len(self)) + b''.join((
-            le_u256enc(length) + b64dec(id)
+        return self.get_count().to_bytes(32, 'little') + b''.join((
+            int(length).to_bytes(32, 'little') + b64dec(id)
             for id, length in self.length_by_id.items()
         ))
 
     @classmethod
     def from_tags_stream(cls, tags, stream):
-        fmt = tags[b'Bundle-Format']
+        fmt = tags_to_dict(tags).get(b'Bundle-Format')
         if fmt == b'json':
             return cls.fromjson(json.load(stream))
         elif fmt == b'binary':
@@ -67,7 +65,7 @@ class ANS104BundleHeader:
     def fromjson(cls, json):
         dataitems = (DataItem.fromjson(item) for item in json['items'])
         return cls({
-            dataitem.id: dataitem.get_len_bytes()
+            dataitem.header.id: dataitem.get_len_bytes()
             for dataitem in dataitems
         }, version=1)
 
@@ -78,10 +76,10 @@ class ANS104BundleHeader:
 
     @classmethod
     def fromstream(cls, stream):
-        entryct = le_u256dec(stream.read(32))
+        entryct = int.from_bytes(stream.read(32), 'little')
 
         length_id_pairs = (
-            (le_u256dec(stream.read(32)), b64enc(stream.read(32)))
+            (int.from_bytes(stream.read(32), 'little'), b64enc(stream.read(32)))
             for idx in range(entryct)
         )
 
@@ -120,7 +118,7 @@ class ANS104DataItemHeader:
             return None
 
     @signature.setter
-    def set_signature(self, signature):
+    def signature(self, signature):
         self.raw_signature = b64dec(signature)
 
     @property
@@ -168,16 +166,11 @@ class ANS104DataItemHeader:
     @property
     def tags(self):
         stream = io.BytesIO(self.raw_tags)
-        return fastavro.schemaless_reader(stream, ANS104_TAGS_SCHEMA_fastavro)
+        return self.tagsfromstream(stream)
 
     @tags.setter
     def tags(self, tags):
-        stream = io.BytesIO()
-        fastavro.schemaless_writer(stream, ANS104_TAGS_SCHEMA_fastavro, [
-            normalize_tag(tag) for tag in tags
-        ])
-        self.raw_tags = stream.getvalue()
-        
+        self.raw_tags = self.tagstobytes(tags)
 
     def tojson(self):
         return {
@@ -301,7 +294,7 @@ class ANS104DataItemHeader:
         if raw_tags_len > 0:
             raw_tags = stream.read(raw_tags_len)
             raw_tags_stream = io.BytesIO(raw_tags)
-            tags = fastavro.schemaless_reader(raw_tags_stream, ANS104_TAGS_SCHEMA_fastavro)
+            tags = cls.tagsfromstream(raw_tags_stream)
             offset += raw_tags_len
             if raw_tags_stream.tell() != raw_tags_len or len(tags) != tags_len:
                 raise Exception(f'incorrect tags length')
@@ -323,6 +316,72 @@ class ANS104DataItemHeader:
         assert offset == result.get_len_bytes()
         return result
 
+    @staticmethod
+    def tagsfromstream(stream):
+        def avrolongdec(stream):
+            # varint encoding is 7-bit little-endian where the 8th bit indicates whether another byte follows
+            zigzag = 0
+            bits = 0
+            while True:
+                byte = stream.read(1)[0]
+                zigzag |= (byte & 0x7f) << bits
+                if byte & 0x80:
+                    bits += 7
+                else:
+                    break
+            # zigzag encoding has the sign bit in the 1s place
+            if zigzag & 1:
+                zigzag = ~zigzag
+            return zigzag >> 1
+        def avrobytesdec(stream):
+            count = avrolongdec(stream)
+            assert count >= 0
+            result = stream.read(count)
+            assert len(result) == count
+            return result
+        tags = []
+        while True:
+            block_items = avrolongdec(stream)
+            if block_items == 0:
+                break
+            elif block_items < 0:
+                block_bytes = avrolongdec(stream)
+                block_items = -block_items
+            for block_idx in range(block_items):    
+                name = avrobytesdec(stream)
+                value = avrobytesdec(stream)
+                tags.append(create_tag(name, value, True))
+        return tags
+
+    @staticmethod
+    def tagstobytes(tags):
+        def avrolongenc(num):
+            # zigzag encoding moves the sign bit to the 1s place
+            zigzag = num << 1
+            if zigzag < 0:
+                zigzag = ~zigzag
+            # varint encoding is 7-bit little-endian where the 8th bit indicates whether another byte follows
+            varint = bytearray()
+            while True:
+                byte = zigzag & 0x7f
+                zigzag >>= 7
+                if zigzag:
+                    varint.append(byte | 0x80)
+                else:
+                    varint.append(byte)
+                    return varint
+        def avrobytesenc(data):
+            return avrolongenc(len(data)) + data
+        return b''.join((
+            avrolongenc(len(tags)),
+            *(
+                avrobytesenc(tag[key])
+                for tag in (normalize_tag(tag) for tag in tags)
+                for key in ("name", "value")
+            ),
+            b'\0'
+        ))
+
 class DataItem:
     def __init__(self, header = None, data = b'', version = 2):
         self.data = data
@@ -436,7 +495,7 @@ class Bundle:
     @property
     def header(self):
         return ANS104BundleHeader({
-            item.id: item.get_len_bytes()
+            item.header.id: item.get_len_bytes()
             for item in self.dataitems
         })
 
@@ -449,7 +508,7 @@ class Bundle:
         }
 
     def tobytes(self):
-        return self.header + b''.join(item.tobytes() for item in self.dataitems)
+        return self.header.tobytes() + b''.join(item.tobytes() for item in self.dataitems)
     
     @classmethod
     def fromjson(cls, json):
@@ -459,7 +518,7 @@ class Bundle:
     @classmethod
     def frombytes(cls, bytes):
         stream = io.BytesIO(bytes)
-        return self.fromstream(stream)
+        return cls.fromstream(stream)
 
     @classmethod
     def fromstream(cls, stream):
