@@ -1,74 +1,209 @@
-import erlang
-import requests
-import json
+import io
+import threading
+import time
 
-from . import DEFAULT_API_URL, logger, ArweaveException, ArweaveNetworkException
-from .utils import response_stream_to_file_object, b64dec, arbindec
+import erlang
+import json
+import requests
+
+from . import DEFAULT_API_URL, DEFAULT_REQUESTS_PER_MINUTE_LIMIT, logger, ArweaveException, ArweaveNetworkException
+from .stream import PeerStream, GatewayStream
+from .utils import b64dec, arbindec
+
+def binary_to_term(b):
+    # arweave.live seems to replace nonascii chars with this sequence :/
+    # occasionally things will work if it's turned back into the etf byte
+    if b[:3] == b'\xef\xbf\xbd':
+        b = b.replace(b'\xef\xbf\xbd', b'\x83')
+    return erlang.binary_to_term(b)
 
 class HTTPClient:
-    def __init__(self, api_url, timeout = None, retries = 5):
+    def __init__(self, api_url, timeout = None, retries = 10, outgoing_connections = DEFAULT_REQUESTS_PER_MINUTE_LIMIT, requests_per_period = DEFAULT_REQUESTS_PER_MINUTE_LIMIT, period_sec = 60):
         self.api_url = api_url
         self.session = requests.Session()
+        self.max_outgoing_connections = outgoing_connections
+        self.outgoing_connection_semaphore = threading.BoundedSemaphore(outgoing_connections)
+        self.rate_limit_lock = threading.Lock()
+        self.requests_per_period = requests_per_period
+        self.ratelimited_requests = 0
+        self.period_sec = period_sec
+        self.req_history = []
         max_retries = requests.adapters.Retry(total=retries, backoff_factor=0.1, status_forcelist=[500,502,503,504]) # from so
-        self.session.mount('http://', requests.adapters.HTTPAdapter(max_retries=max_retries))
-        self.session.mount('https://', requests.adapters.HTTPAdapter(max_retries=max_retries))
+        adapter = requests.adapters.HTTPAdapter(
+            pool_connections = outgoing_connections,
+            pool_maxsize = outgoing_connections,
+            max_retries = max_retries,
+            pool_block = True
+        )
+        self.session.mount('http://', adapter)
+        self.session.mount('https://', adapter)
         self.timeout = timeout
+    def __del__(self):
+        self.session.close()
+
+    def ratelimited(self):
+        if self.requests_per_period is None:
+            return False
+        with self.rate_limit_lock:
+            now = time.time()
+            queued_requests = 0
+            for idx, then in enumerate(self.req_history):
+                if then + self.period_sec >= now:
+                    queued_requests_idx = idx
+                    queued_requests = len(self.req_history) - queued_requests_idx
+                    break
+            return queued_requests + self.ratelimited_requests >= self.requests_per_period
+
+    def ratelimit_suggested(self):
+        if self.requests_per_period is None:
+            return False
+        if len(self.req_history) == 0:
+            return False
+        return (time.time() - self.req_history[-1] < self.requests_per_period / period_sec) or self.ratelimited
+
+    def _ratelimit_prologue(self):
+        if self.requests_per_period is None:
+            return
+        with self.rate_limit_lock:
+            now = time.time()
+            queued_requests = 0
+            #if len(self.req_history) >= 3600:
+            #    import pdb; pdb.set_trace()
+            for idx, then in enumerate(self.req_history):
+                if then + self.period_sec >= now:
+                    queued_requests_idx = idx
+                    queued_requests = len(self.req_history) - queued_requests_idx
+                    break
+            if queued_requests + self.ratelimited_requests < self.requests_per_period:
+                #if len(self.req_history) >= self.requests_per_period:
+                #    import pdb; pdb.set_trace()
+                self.req_history.append(now)
+                return
+        #print(f'{self.api_url}: too many requests in prologue')
+        self.on_too_many_requests()
+        with self.rate_limit_lock:
+            now = time.time()
+            if len(self.req_history) >= self.requests_per_period:
+                duration = self.req_history[-self.requests_per_period+1] + self.period_sec - now
+                if duration > 0:
+                    logger.info(f'Sleeping for {int(duration*100)/100}s to respect ratelimit of {self.requests_per_period}req/{self.period_sec}s ...')
+                    time.sleep(duration)
+                    #import pdb; pdb.set_trace()
+                    logger.info(f'Done sleeping for {int(duration*100)/100}s to respect ratelimit of {self.requests_per_period}req/{self.period_sec}s .')
+        return self._ratelimit_prologue()
+
+    def _ratelimit_epilogue(self, success = True):
+        if self.requests_per_period is None:
+            return
+        if success:
+            with self.rate_limit_lock:
+                self.ratelimited_requests = 0
+                now = time.time()
+                for req_idx, req_time in enumerate(self.req_history):
+                    if req_time + self.period_sec > now:
+                        self.req_history = self.req_history[req_idx:]
+                        break
+        else:
+            with self.rate_limit_lock:
+                self.ratelimited_requests += 1
+                now = time.time()
+                #import pdb; pdb.set_trace()
+                if len(self.req_history):
+                    self.period_sec = max(self.period_sec, now - self.req_history[0])
+                if len(self.req_history) - self.ratelimited_requests <= self.requests_per_period:
+                    self.requests_per_period = max(1, len(self.req_history) - self.ratelimited_requests)
+                logger.info(f'Rate limit hit. Dropped rate to {self.requests_per_period}/{self.period_sec}s.')
+            self.on_too_many_requests()
+
+    # _get and _post should just call a _request function to share code
+    def _request(self, *params, **request_kwparams):
+        if len(params) and params[-1][0] == '?':
+            url = self.api_url + '/' + '/'.join(params[:-1]) + params[1]
+        else:
+            url = self.api_url + '/' + '/'.join(params)
+
+        while True:
+            self._ratelimit_prologue()
+            response = None
+            try:
+                if not self.outgoing_connection_semaphore.acquire(blocking=False):
+                    self.on_too_many_connections()
+                    logger.info(f'Waiting for connection count limit semaphore to drain...')
+                    self.outgoing_connection_semaphore.acquire()
+                try:
+                    response = self.session.request(**{'url': url, 'timeout': self.timeout, **request_kwparams})
+                finally:
+                    self.outgoing_connection_semaphore.release()
+
+                if response.status_code == 400:
+                    try:
+                        msg = response.json()['error']
+                    except:
+                        msg = response.text
+                    raise ArweaveException(msg)
+
+                response.raise_for_status()
+                if int(response.headers.get('content-length', 1)) == 0:
+                    raise ArweaveException(f'Empty response from {url}')
+                self._ratelimit_epilogue(True)
+                return response
+            except requests.exceptions.RequestException as exc:
+                text = '' if response is None else response.text
+                status_code = 0 if response is None else response.status_code
+                if status_code == 429:
+                    # too many requests
+                    self._ratelimit_epilogue(False)
+                    self.on_too_many_requests()
+                    continue
+                logger.error('{}\n{}\n\n{}'.format(exc, text, request_kwparams))
+                if status_code == 520:
+                    # cloudfront broke
+                    self._ratelimit_epilogue(True)
+                    continue
+                elif status_code == 502:
+                    # cloudflare broke
+                    self._ratelimit_epilogue(True)
+                    continue
+                self.on_network_exception(text, status_code, exc, response)
+                raise ArweaveNetworkException(text, status_code, exc, response)
+            except:
+                self._ratelimit_epilogue(True)
+                raise
 
     def _get(self, *params, **request_kwparams):
-        if len(params) and params[-1][0] == '?':
-            url = self.api_url + '/' + '/'.join(params[:-1]) + params[1]
-        else:
-            url = self.api_url + '/' + '/'.join(params)
-
-        response = self.session.request(**{'method': 'GET', 'url': url, 'timeout': self.timeout, **request_kwparams})
-
-        try:
-            response.raise_for_status()
-            if int(response.headers.get('content-length', 1)) == 0:
-                raise ArweaveException('Empty response.')
-            if response.status_code not in (200, 206):
-                raise ArweaveException(response.text)
-            return response
-        except requests.exceptions.RequestException as exc:
-            logger.error(response.text)
-            raise ArweaveNetworkException(response.text, exc, response)
+        return self._request(*params, **{'method': 'GET', **request_kwparams})
 
     def _post(self, data, *params, headers = {}, **request_kwparams):
-        if len(params) and params[-1][0] == '?':
-            url = self.api_url + '/' + '/'.join(params[:-1]) + params[1]
-        else:
-            url = self.api_url + '/' + '/'.join(params)
-
         headers = {**headers}
 
         if type(data) is dict:
-            if type(data) is dict:
-                headers.setdefault('Content-Type', 'application/json')
-            response = self.session.request(**{'method': 'POST', 'url': url, 'json': data, 'headers': headers, 'timeout': self.timeout, **request_kwparams})
+            headers.setdefault('Content-Type', 'application/json')
+            data_key = 'json'
         else:
             if isinstance(data, (bytes, bytearray)):
                 headers.setdefault('Content-Type', 'application/octet-stream')
             else:
                 headers.setdefault('Content-Type', 'text/plain')
-            response = self.session.request(**{'method': 'POST', 'url': url, 'data': data, 'headers': headers, 'timeout': self.timeout, **request_kwparams})
+            data_key = 'data'
 
-        # logger.debug('{}\n\n{}'.format(response.text, data))
+        return self._request(*params, **{'method': 'POST', 'headers': headers, **{data_key: data}, **request_kwparams})
 
-        try:
-            response.raise_for_status()
-            # logger.debug('RESPONSE 200: {}'.format(response.text))
-            return response
-        except requests.exceptions.RequestException as exc:
-            logger.error('{}\n\n{}'.format(response.text, data))
-            raise ArweaveNetworkException(response.text, exc, response)
+    def on_network_exception(self, text, code, exception, response):
+        raise ArweaveNetworkException(text, code, exception, response)
+
+    def on_too_many_connections(self):
+        pass
+
+    def on_too_many_requests(self):
+        return self.on_too_many_connections()
 
 class Peer(HTTPClient):
     # peer api [incomplete]:
     # - https://docs.arweave.org/developers/server/http-api
     # - https://github.com/ArweaveTeam/arweave/blob/master/apps/arweave/src/ar_http_iface_middleware.erl#L132
     # - https://github.com/ArweaveTeam/arweave/blob/master/apps/arweave/src/ar_http_iface_client.erl
-    def __init__(self, api_url = DEFAULT_API_URL, timeout = None, retries = 5):
-        super().__init__(api_url, timeout, retries)
+    def __init__(self, api_url = DEFAULT_API_URL, timeout = None, retries = 5, outgoing_connections = DEFAULT_REQUESTS_PER_MINUTE_LIMIT, requests_per_period = DEFAULT_REQUESTS_PER_MINUTE_LIMIT, period_sec = 60):
+        super().__init__(api_url, timeout, retries, outgoing_connections, requests_per_period, period_sec)
 
     def info(self):
         '''
@@ -106,12 +241,12 @@ class Peer(HTTPClient):
             for height in range(self.height(), 0, -1):
                 tx = self.block_height(height)['txs'][0]
                 try:
-                    stream = self.stream(tx, range = (0,4))
+                    stream = self.gateway_stream(tx, range = (0,4))
                     data = stream.read(8)
                     if len(data) == 8:
                         partial_data = False
                     else:
-                        stream = self.stream(tx)
+                        stream = self.gateway_stream(tx)
                         data = stream.read(8)
                         if len(data) == 8:
                             partial_data = True
@@ -126,7 +261,7 @@ class Peer(HTTPClient):
 
         def peers():
             health = status.get('health')
-            if health is not None:
+            if health is not None and 'origins' in health:
                 status['peers'] = [origin['endpoint'].split('://',1)[1] for origin in health['origins']]
                 raise ArweaveNetworkException()
             else:
@@ -231,7 +366,12 @@ class Peer(HTTPClient):
 
     def tx2(self, txid):
         '''Return a binary-encoded transaction.'''
-        response = self._get('tx2', txid)
+        response = self._get('tx2', txid)#, stream = True)
+        #with response_stream_to_file_object(response) as stream:
+        #    size = int.from_bytes(stream.read(3), 'big')
+        #    format = stream.read(1)[0]
+        #    txid_raw = stream.read(32)
+        #    last_tx = arbindec(stream, 
         return response.content
 
     def unconfirmed_tx(self, txid):
@@ -301,7 +441,7 @@ class Peer(HTTPClient):
         every value - the percentage of data synced in the reported bucket.
         '''
         response = self._get('sync_buckets')
-        return erlang.binary_to_term(response.content)
+        return binary_to_term(response.content)
 
     def data_sync_record(self, start = None, limit = None, format = 'etf'):
         '''
@@ -335,9 +475,10 @@ class Peer(HTTPClient):
             except json.decoder.JSONDecodeError:
                 # some proxies, such as arweave.net 2022-05, ignore the header and return etf
                 pass
-        intervals = erlang.binary_to_term(response.content)
+        intervals = binary_to_term(response.content)
+                
         intervals = [
-            (int.from_bytes(left.value, 32, 'big'), int.from_bytes(right.value, 32, 'big'))
+            (int.from_bytes(left.value, 'big'), int.from_bytes(right.value, 'big'))
             for left, right in intervals
         ]
         return intervals
@@ -347,7 +488,7 @@ class Peer(HTTPClient):
         Returns the data chunk containing 1-based offset,
         using json, base64 encoded network transmission.
 
-        {packing} := { 'unpacked' | 'spora_2_5' | 'any' }
+        {packing} := { 'unpacked' | 'spora_2_5' | 'spora_2_6_<address>' | 'any' }
 
         {
             "tx_path",
@@ -364,39 +505,14 @@ class Peer(HTTPClient):
             headers['x-bucket-based-offset'] = '1'
 
         response = self._get('chunk', str(offset), headers=headers)
-        result = response.json()
-        for key in ('tx_path', 'data_path', 'chunk'):
-            result[key] = b64dec(result[key])
-        return result
+        return response.json()
 
     def chunk2(self, offset, packing = 'unpacked', bucket_based_offset = False):
         '''
         Returns the data chunk containing 1-based offset,
-        using raw binary network transmission.
+        using a raw binary format for the fields.
 
-        {packing} := { 'unpacked' | 'spora_2_5' | 'any' }
-
-        The raw data format from the endpoint is b[
-            chunk_size      3 bytes, big-endian
-            chunk           chunk_size bytes
-
-            txpath_size     3 bytes, big-endian
-            txpath          txpath_size bytes
-
-            datapath_size   3 bytes, big-endian
-            datapath        datapath_size bytes
-
-            packing2_size   1 byte
-            packing2        packing2_size bytes
-        ]
-
-        This function parses it to return the same format as chunk():
-        {
-            "chunk": b"<chunk>",
-            "tx_path": b"<txpath>",
-            "data_path"; b"<datapath>",
-            "packing": "packing"
-        }
+        {packing} := { 'unpacked' | 'spora_2_5' | 'spora_2_6_<address>' | 'any' }
         '''
 
         headers = {
@@ -405,17 +521,8 @@ class Peer(HTTPClient):
         if bucket_based_offset:
             headers['x-bucket-based-offset'] = '1'
 
-        response = self._get('chunk2', str(offset), headers=headers, stream = True)
-        with response_stream_to_file_object(response) as stream:
-            result = {}
-            result['chunk'] = arbindec(stream, 24)
-            result['tx_path'] = arbindec(stream, 24)
-            result['data_path'] = arbindec(stream, 24)
-            result['packing'] = arbindec(stream, 8).decode()
-            extra = stream.read()
-            if extra:
-                result['_extra'] = extra
-            return result
+        response = self._get('chunk2', str(offset), headers=headers)
+        return response.content
 
     def chunk_size(self, offset, packing = 'unpacked', bucket_based_offset = False):
         '''Returns the size of the data chunk containing 1-based offset.'''
@@ -452,6 +559,10 @@ class Peer(HTTPClient):
         return result
 
     def send_chunk(self, json_data):
+        # NOTE: this can take two headers
+        # arweave-data-root: 43 encoded bytes
+        # arweave-data-size: integer
+        # i'm guessing these are used to quickly discard data that does not match the server
         '''
         Upload Data Chunks
 
@@ -464,8 +575,15 @@ class Peer(HTTPClient):
           "offset": "<a number from [start_offset, start_offset + chunk size), relative to other chunks>"
         }
         '''
-        response = self._post(json_data, 'chunk')
-        return response.json()
+        response = self._post(
+            json_data,
+            'chunk',
+            headers={
+                'arweave-data-root': json_data['data_root'],
+                'arweave-data-size': str(json_data['data_size'])
+            }
+        )
+        return response.text # OK
 
     def block_announcement(self, block_announcement):
         '''
@@ -476,14 +594,20 @@ class Peer(HTTPClient):
         response = self._post(block_announcement, 'block_announcement')
         return response.json()
 
-    def send_block(self, block):
+    def send_block(self, block, arweave_recall_byte : int = None):
         '''Accept a JSON-encoded block with Base64Url encoded fields.'''
-        response = self._post(block, 'block')
+        headers = {}
+        if arweave_recall_byte is not None:
+            headers['arweave-recall-byte'] = str(arweave_recall_byte)
+        response = self._post(block, 'block', headers=headers)
         return response.text # OK
 
-    def send_block2(self, block):
+    def send_block2(self, block, arweave_recall_byte : int = None):
         '''Accept a binary-encoded block.'''
-        response = self._post(block, 'block2')
+        headers = {}
+        if arweave_recall_byte is not None:
+            headers['arweave-recall-byte'] = str(arweave_recall_byte)
+        response = self._post(block, 'block2', headers=headers)
         return response.text # OK
 
     def wallet(self, secret):
@@ -661,7 +785,7 @@ class Peer(HTTPClient):
         Return transaction identifiers (hashes), optionally starting from the earliest_tx,
         for the wallet specified via wallet_address.
         '''
-        if earlieset_tx is not None:
+        if earliest_tx is not None:
             response = self._get('wallet', wallet_address, 'txs', earliest_tx)
         else:
             response = self._get('wallet', wallet_address, 'txs')
@@ -682,6 +806,9 @@ class Peer(HTTPClient):
         '''
         Return the JSON-encoded block or field of a block with the given hash.
 
+        field :: nonce | previous_block | timestamp | last_retarget | diff | height | hash |
+        indep_hash | txs | hash_list | wallet_list | reward_addr | tags | reward_pool
+
         {
             "nonce",
             "previous_block",
@@ -714,7 +841,12 @@ class Peer(HTTPClient):
         return response.json()
 
     def block_height(self, height, field = None):
-        '''Return the JSON-encoded block or field of a block with the given height.'''
+        '''
+        Return the JSON-encoded block or field of a block with the given height.
+
+        field :: nonce | previous_block | timestamp | last_retarget | diff | height | hash |
+        indep_hash | txs | hash_list | wallet_list | reward_addr | tags | reward_pool
+        '''
         if field is not None:
             response = self._get('block/height', str(height), field)
         else:
@@ -766,6 +898,7 @@ class Peer(HTTPClient):
         return response.json()
 
     def block(self, height_or_hash):
+        '''A convenience method that hands off to block_height or block_hash.'''
         if type(height_or_hash) is int:
             return self.block_height(height_or_hash)
         elif not height_or_hash :
@@ -774,6 +907,7 @@ class Peer(HTTPClient):
             return self.block_hash(height_or_hash)
 
     def block2(self, height_or_hash):
+        '''A convenience method that hands off to block2_height or block2_hash.'''
         if type(height_or_hash) is int:
             return self.block2_height(height_or_hash)
         else:
@@ -800,7 +934,10 @@ class Peer(HTTPClient):
                         tag[key] = b64dec(tag[key].encode())
                 return tags
             else:
-                return response.json()
+                try:
+                    return response.json()
+                except json.decoder.JSONDecodeError:
+                    return response.text
 
     def tx_id(self, hash):
         '''Return transaction id.'''
@@ -873,13 +1010,26 @@ class Peer(HTTPClient):
 
         return response.content
 
-    def stream(self, txid, ext = '', range = None):
+    def stream(self, txid, range = None):
+        try:
+            return self.peer_stream(txid, range=range)
+        except Exception:
+            if reupload(self, txid):
+                return self.gateway_stream(txid, range=range)
+            else:
+                raise
+
+    def gateway_stream(self, txid, ext ='', range = None):
         if range is not None:
-            headers = {'Range':f'bytes={range[0]}-{range[1]}'}
+            return GatewayStream.from_txid(self, txid, range[0], range[1] - range[0])
         else:
-            headers = {}
-        response = self._get(txid + ext, headers = headers, stream = True)
-        return response_stream_to_file_object(response)
+            return GatewayStream.from_txid(self, txid)
+
+    def peer_stream(self, txid, range = None):
+        if range is not None:
+            return io.BufferedReader(PeerStream.from_txid(self, txid, range[0], range[1]-range[0]), 0x40000)
+        else:
+            return io.BufferedReader(PeerStream.from_txid(self, txid), 0x40000)
 
     # below are used with https://github.com/ar-io/arweave-gateway
 
@@ -964,3 +1114,28 @@ class Peer(HTTPClient):
     def cache_jobs(self):
         response = self._get('cache', 'jobs')
         return response.json()
+
+from ar.utils.merkle import compute_root_hash, generate_transaction_chunks
+from ar.utils import b64enc
+def reupload(peer, tx):
+    stream = peer.gateway_stream(tx)
+
+    chunks = generate_transaction_chunks(stream)
+    stream = peer.gateway_stream(tx)
+    if chunks['data_root'] != peer.tx_data_root(tx):
+        logger.error(f'{peer.api_url}: Data for {tx} mismatches generated root.')
+        return False
+    logger.warning(f'uhh trying to reupload {tx}')
+    offset = 0
+    for proof, chunk in zip(chunks['proofs'], chunks['chunks']):
+        chunk_size = chunk.data_size
+        chunk = {
+            'data_root': chunks['data_root'],
+            'data_size': str(chunks['chunks'][-1].max_byte_range),
+            'data_path': b64enc(proof.proof),
+            'offset': str(proof.offset),
+            'chunk': b64enc(stream.read(chunk_size))
+        }
+        peer.send_chunk(chunk)
+        offset+=chunk_size
+    return True
