@@ -15,9 +15,8 @@
 import hashlib
 import struct
 import functools
-from jose.utils import base64url_encode, base64url_decode
 from .file_io import read_file_chunks
-from . import concat_buffers
+from . import concat_buffers, b64enc, b64dec
 from json import JSONEncoder
 
 CHUNK_SIZE = 256 * 1024
@@ -32,30 +31,160 @@ class NodeTypeException(Exception):
 
 
 class Node:
-    def __init__(self, id='', type=None, byte_range=0, max_byte_range=0):
-        self.id = id
+    def __init__(self, id_raw=b'', type=None, byte_range=0, max_byte_range=0, parent=None):
+        self.id_raw = id_raw
         self.type = type
         self.byte_range = byte_range
         self.max_byte_range = max_byte_range
+        self.parent = parent
+
+    # I patched binary coding of data paths in here, but it may go better
+    # elsewhere. Haven't reviewed most of these things.
+    @classmethod
+    def frombytes(cls, bytes, min_byte_range = 0, max_byte_range = None):
+        import io; stream = io.BytesIO(bytes)
+        return cls.fromstream(stream, len(bytes), min_byte_range)
+    @staticmethod
+    def fromstream(stream, length, min_byte_range = 0, max_byte_range = None):
+        branches_raw = []
+        branches = []
+        remaining = length
+
+        while remaining > HASH_SIZE + NOTE_SIZE:
+            raw = stream.read(2*HASH_SIZE + NOTE_SIZE)
+            left_raw, right_raw, midpoint_raw = (
+                raw[:HASH_SIZE],
+                raw[HASH_SIZE:2*HASH_SIZE],
+                raw[2*HASH_SIZE:]
+
+            )
+            midpoint = int.from_bytes(midpoint_raw, 'big')
+            id_raw = hash_raw([
+                hash_raw(left_raw),
+                hash_raw(right_raw),
+                hash_raw(midpoint_raw)
+            ])
+            branches_raw.append((left_raw, right_raw))
+            branches.append(BranchNode(
+                    id_raw = id_raw,
+                    byte_range = midpoint,
+                    max_byte_range = None,
+                    parent = branches[-1] if len(branches) else None
+            ))
+            remaining -= 2*HASH_SIZE + NOTE_SIZE
+
+        leaf_raw = stream.read(HASH_SIZE + NOTE_SIZE)
+        leaf_data_hash = leaf_raw[:HASH_SIZE]
+        leaf_max_byte_range = buffer_to_int(leaf_raw[HASH_SIZE:])
+        branches_raw.append((leaf_data_hash,))
+        branches.append(LeafNode(
+            max_byte_range = leaf_max_byte_range,
+            data_hash = leaf_data_hash,
+            min_byte_range = None,
+            parent = branches[-1] if len(branches) else None
+        ))
+
+        branches[0].min_byte_range = min_byte_range
+        if branches[0].max_byte_range is None:
+            branches[0].max_byte_range = max_byte_range
+
+        leaf_min_byte_range = min_byte_range
+
+        for branch, next_branch, (l_hash, r_hash) in zip(branches[:-1], branches[1:], branches_raw):
+            midpoint = branch.byte_range
+            if leaf_max_byte_range <= midpoint:
+                assert next_branch.id_raw == l_hash
+                if type(next_branch) is BranchNode:
+                    assert next_branch.byte_range <= midpoint
+                next_branch.min_byte_range = branch.min_byte_range
+                if type(next_branch) is LeafNode:
+                    assert next_branch.max_byte_range == midpoint
+                next_branch.max_byte_range = midpoint
+                branch.left_child = next_branch
+                branch.right_child = type(next_branch)(
+                    id_raw = r_hash,
+                    min_byte_range = midpoint,
+                    max_byte_range = branch.max_byte_range if branch.max_byte_range is not None else 0,
+                    byte_range = None,
+                    parent = branch,
+                    data_hash = b''
+                )
+                branch.right_child.id_raw = r_hash
+            else: # leaf_max_byte_range > midpoint
+                assert next_branch.id_raw == r_hash
+                if type(next_branch) is BranchNode:
+                    assert next_branch.byte_range >= midpoint # note that branches can be zero-sized
+                next_branch.min_byte_range = midpoint
+                if type(next_branch) is LeafNode:
+                    if branch.max_byte_range is not None:
+                        assert next_branch.max_byte_range == branch.max_byte_range
+                else:
+                    next_branch.max_byte_range = branch.max_byte_range
+                branch.left_child = type(next_branch)(
+                    id_raw = l_hash,
+                    min_byte_range = branch.min_byte_range,
+                    max_byte_range = midpoint,
+                    byte_range = None,
+                    parent = branch,
+                    data_hash = b''
+                )
+                branch.left_child.id_raw = l_hash
+                branch.right_child = next_branch
+        branches[0].leaf = branches[-1]
+        branches[-1].root = branches[0]
+        return branches
 
 
 class BranchNode(Node):
     def __init__(self, *args, **kwargs):
-        super(BranchNode, self).__init__(id=kwargs['id'], max_byte_range=kwargs['max_byte_range'],
-                                         byte_range=kwargs['byte_range'])
+        super(BranchNode, self).__init__(id_raw=kwargs['id_raw'], max_byte_range=kwargs['max_byte_range'],
+                                         byte_range=kwargs['byte_range'], parent=kwargs['parent'])
         self.type = 'branch'
         self.left_child = kwargs.get('left_child', None)
         self.right_child = kwargs.get('right_child', None)
+        if self.left_child is not None:
+            self.left_child.parent = self
+        if self.right_child is not None:
+            self.right_child.parent = self
+
+    def add(self, node):
+        if node.max_byte_range <= self.byte_range:
+            assert self.left_child is None
+            self.left_child = node
+        else:
+            assert self.right_child is None
+            self.right_child = node
+
+    def tobytes(self):
+        if self.byte_range is not None:
+            return b''.join((
+                self.left_child.id_raw,
+                self.right_child.id_raw,
+                int_to_buffer(self.byte_range),
+                self.left_child.tobytes(),
+                self.right_child.tobytes()
+            ))
+        else:
+            return b''
 
 
 class LeafNode(Node):
     def __init__(self, *args, **kwargs):
-        super(LeafNode, self).__init__(max_byte_range=kwargs['max_byte_range'])
-        self.data_hash = kwargs['data_hash']
+        super(LeafNode, self).__init__(max_byte_range=kwargs['max_byte_range'],parent=kwargs['parent'])
+        self.data_hash = kwargs['data_hash'] # raw bytes
         self.min_byte_range = kwargs['min_byte_range']
-        self.id = hash([hash(self.data_hash), hash(int_to_buffer(self.max_byte_range))])
+        self.id_raw = hash_raw([hash_raw(self.data_hash), hash_raw(int_to_buffer(self.max_byte_range))])
         self.type = 'leaf'
 
+    def tobytes(self):
+        if self.data_hash:
+            return b''.join((
+                self.data_hash,
+                int_to_buffer(self.max_byte_range)
+            ))
+        else:
+            return b''
+
 
 class TaggedChunk:
     def __init__(self, tc_id, end):
@@ -73,7 +202,7 @@ class Chunk:
     def to_dict(self):
         print('boom!')
         return {
-            'dataHash': base64url_encode(self.data_hash).decode(),
+            'dataHash': b64enc(self.data_hash).decode(),
             'maxByteRange': self.max_byte_range,
             'minByteRange': self.min_byte_range
         }
@@ -93,7 +222,7 @@ class Proof:
     def to_dict(self):
         return {
             'offset': self.offset,
-            'proof': base64url_encode(self.proof).decode()
+            'proof': b64enc(self.proof).decode()
         }
 
 
@@ -119,7 +248,7 @@ def chunk_data(file_handler):
     cursor = 0
 
     for chunk in read_file_chunks(file_handler, MAX_CHUNK_SIZE):
-        data_hash = hashlib.sha256(chunk).digest()
+        data_hash = hash_raw(chunk)
 
         cursor += len(chunk)
 
@@ -146,7 +275,8 @@ def generate_leaves(chunks):
         LeafNode(
             data_hash=chunk.data_hash,
             min_byte_range=chunk.min_byte_range,
-            max_byte_range=chunk.max_byte_range
+            max_byte_range=chunk.max_byte_range,
+            parent=None
         )
         for chunk in chunks
     ]
@@ -206,7 +336,7 @@ def generate_transaction_chunks(file_handler):
         proofs = proofs[:-1]
 
     return {
-        'data_root': root.id,
+        'data_root': b64enc(root.id_raw),
         'chunks': chunks,
         'proofs': proofs
     }
@@ -249,8 +379,8 @@ def resolve_branch_proofs(node, proof=b'', depth=0):
     if node.type == 'branch':
         partial_proof = concat_buffers([
             proof,
-            node.left_child.id,
-            node.right_child.id,
+            node.left_child.id_raw,
+            node.right_child.id_raw,
             int_to_buffer(node.byte_range)
         ])
 
@@ -267,104 +397,76 @@ def hash_branch(left, right=None):
         return left
 
     return BranchNode(
-        id=hash(
+        id_raw=hash_raw(
             [
-                hash(left.id),
-                hash(right.id),
-                hash(int_to_buffer(left.max_byte_range))
+                hash_raw(left.id_raw),
+                hash_raw(right.id_raw),
+                hash_raw(int_to_buffer(left.max_byte_range))
             ]
         ),
         byte_range=left.max_byte_range,
         max_byte_range=right.max_byte_range,
         left_child=left,
-        right_child=right
+        right_child=right,
+        parent=None
     )
 
 
 def hash_leaf(data, note):
     return HashNode(
-        hash([hash(data), hash(note_to_buffer(note))]),
+        hash([hash(data), hash(int_to_buffer(note))]),
         note
     )
 
 
-def hash(data):
+def hash_raw(data):
     if type(data) == list:
-        byte_str = b''
-        for line in data:
-            byte_str += line
-
-        data = byte_str
+        data = b''.join(data)
 
     digest = hashlib.sha256(data).digest()
-    b64_str = base64url_encode(digest)
     return digest
 
 
-def note_to_buffer(note):
-    buffer = b'\x00' * NOTE_SIZE
-    buffer = bytearray(buffer)
-
-    for i in range(NOTE_SIZE - 1, 0, -1):
-        if i > 0:
-            if note > 0:
-                buffer[i] = note.to_bytes(4, byteorder='big')[-1]
-                note = note >> 8
-            else:
-                break
-        else:
-            break
-
-    return bytes(buffer)
+def hash(data):
+    digest = hash_raw(data)
+    b64_str = b64enc(digest)
+    return digest
 
 
 def int_to_buffer(note):
-    buffer = b'\x00' * NOTE_SIZE
-    buffer = bytearray(buffer)
-
-    for i in range(NOTE_SIZE - 1, 0, -1):
-        byte_val = note % 256
-        buffer[i] = int(byte_val)
-        note = int((note - byte_val) / 256)
-
-    return buffer
+    return int(note).to_bytes(NOTE_SIZE, 'big')
 
 
 def buffer_to_int(buffer):
-    value = 0
-
-    for byte_val in buffer:
-        value *= 256
-        value += byte_val
-
-    return value
+    assert len(buffer) == NOTE_SIZE
+    return int.from_bytes(buffer, 'big')
 
 
 def array_compare(a, b):
     functools.reduce(lambda x, y: x and y, map(lambda p, q: p == q, a, b), True)
 
 
-def validate_path(id, dest, left_bound, right_bound, path):
+def validate_path(id_raw, dest, left_bound, right_bound, path):
     if right_bound < 0:
         return False
 
     if dest > right_bound:
-        return validate_path(id, 0, right_bound - 1, right_bound, path)
+        return validate_path(id_raw, 0, right_bound - 1, right_bound, path)
 
     if dest < 0:
-        return validate_path(id, 0, 0, right_bound, path)
+        return validate_path(id_raw, 0, 0, right_bound, path)
 
     if len(path) == HASH_SIZE + NOTE_SIZE:
         path_data = path[0:HASH_SIZE]
         path_data_length = len(path_data)
         end_offset_buffer = path[path_data_length:path_data_length + NOTE_SIZE]
 
-        path_data_hash = hash([
-            hash(path_data),
-            hash(end_offset_buffer)
+        path_data_hash = hash_raw([
+            hash_raw(path_data),
+            hash_raw(end_offset_buffer)
         ])
 
-        result = id == path_data_hash
+        result = id_raw == path_data_hash
 
         if result:
             return ValidatedPathResult(right_bound - 1, left_bound, right_bound, right_bound - left_bound)
@@ -387,7 +489,7 @@ def validate_path(id, dest, left_bound, right_bound, path):
         hash(offset_buffer)
     ])
 
-    if id == path_hash:
+    if id_raw == path_hash:
         if dest < offset:
             return validate_path(
                 left,
